//-----------------------------LICENSE NOTICE------------------------------------
//   ____        _           _       _____      _                                
//  |  _ \      | |         ( )     |  __ \    | |                               
//  | |_) | __ _| |__   __ _|/ ___  | |__) |_ _| | __ _  ___ ___                 
//  |  _ < / _` | '_ \ / _` | / __| |  ___/ _` | |/ _` |/ __/ _ |                
//  | |_) | (_| | |_) | (_| | \__ \ | |  | (_| | | (_| | (_|  __/                
//  |____/ \__,_|_.__/ \__,_| |___/ |_|   \__,_|_|\__,_|\___\___|                
//
//  Code & Gfx Copyright (C) 2017 Rafael Castillo (@Azicuetano)
//  Music & Fx Copyright (C) 2017 John McKlain (@elmcklain)
//  This file is part of BABA'S PALACE: an Amstrad CPC Game made with CPCTelera
//
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
//------------------------------------------------------------------------------

// * warning * I think this code is a bit inconsistent due to __z88dk_fastcall...

#include <cpctelera.h>
#include <stdio.h>

#include "../globals.h"
#include "../state/game.h"
#include "../state/menu.h"
#include "../keyboard/keyboard.h"

#include "../levels/levels.h"

#include "entities.h"
#include "../sprites/hero_shaolinkid.h"
#include "../sprites/hero_littleindian.h"
#include "../sprites/atlas_enemies.h"
#include "../sprites/atlas_background.h"
#include "../sprites/stone.h"

#include "../utils/uitoa.h"
#include "../text/text.h"


// Face to
enum {
      F_left = 0
   ,  F_right
} EFaceTo;

// Hero Events
enum {
      EV_stop = 0
   ,  EV_move
   ,  EV_jump
   ,  EV_numEvents
} EEventos;


// Heroes sprites: 
// First dimension is for hero: 0=ShaolinKid and 1=Little Indian.
const TFrame g_frames[2][20] = {
   {
      { F_left, gfx_hero_shaolinkid_00 }   ,  { F_left, gfx_hero_shaolinkid_01 }
   ,  { F_left, gfx_hero_shaolinkid_02 }   ,  { F_left, gfx_hero_shaolinkid_03 }
   ,  { F_left, gfx_hero_shaolinkid_04 }   ,  { F_left, gfx_hero_shaolinkid_05 }
   ,  { F_left, gfx_hero_shaolinkid_06 }   ,  { F_left, gfx_hero_shaolinkid_07 }
   ,  { F_left, gfx_hero_shaolinkid_08 }   ,  { F_left, gfx_hero_shaolinkid_09 }
   ,  { F_left, gfx_hero_shaolinkid_10 }   ,  { F_left, gfx_hero_shaolinkid_11 }
   ,  { F_left, gfx_hero_shaolinkid_12 }   ,  { F_left, gfx_hero_shaolinkid_13 }
   ,  { F_left, gfx_hero_shaolinkid_14 }   ,  { F_left, gfx_hero_shaolinkid_15 }
   ,  { F_left, gfx_hero_shaolinkid_16 }   ,  { F_left, gfx_hero_shaolinkid_17 }
   ,  { F_left, gfx_hero_shaolinkid_18 }   ,  { F_left, gfx_hero_shaolinkid_19 } },
   {
      { F_left, gfx_hero_littleindian_00 }   ,  { F_left, gfx_hero_littleindian_01 }
   ,  { F_left, gfx_hero_littleindian_02 }   ,  { F_left, gfx_hero_littleindian_03 }
   ,  { F_left, gfx_hero_littleindian_04 }   ,  { F_left, gfx_hero_littleindian_05 }
   ,  { F_left, gfx_hero_littleindian_06 }   ,  { F_left, gfx_hero_littleindian_07 }
   ,  { F_left, gfx_hero_littleindian_08 }   ,  { F_left, gfx_hero_littleindian_09 }
   ,  { F_left, gfx_hero_littleindian_10 }   ,  { F_left, gfx_hero_littleindian_11 }
   ,  { F_left, gfx_hero_littleindian_12 }   ,  { F_left, gfx_hero_littleindian_13 }
   ,  { F_left, gfx_hero_littleindian_14 }   ,  { F_left, gfx_hero_littleindian_15 }
   ,  { F_left, gfx_hero_littleindian_16 }   ,  { F_left, gfx_hero_littleindian_17 }
   ,  { F_left, gfx_hero_littleindian_18 }   ,  { F_left, gfx_hero_littleindian_19 }
   }
};


// HEROES ANIMATIONS
// Arrays items: {frame from g_frames, displacement }

#define WALK_FRAMES 8
u8 const anim_walk[WALK_FRAMES][2] = {
      { 3, 1}
   ,  { 3, 1}
   ,  { 4, 1}
   ,  { 4, 1}
   ,  { 5, 1}
   ,  { 5, 1}
   ,  { 6, 1}
   ,  { 6, 1}
};

#define PUNCH_FRAMES  10
u8 const anim_punch[PUNCH_FRAMES][2] = {
      { 2, 0}
   ,  { 2, 0}
   ,  { 2, 0}
   ,  { 2, 0}
   ,  {11, 4}
   ,  {11, 2}
   ,  { 9, 1}
   ,  { 9, 0}
   ,  { 9, 0}
   ,  { 9, 0}
};

#define JUMP_FRAMES 6
u8 const anim_jumpTo[JUMP_FRAMES][2] = {
      { 9, 0}
   ,  { 1, 2}
   ,  { 1, 2}
   ,  { 0, 2}
   ,  {14, 1}
   ,  {13, 0}
};

u8 const anim_jumpFrom[JUMP_FRAMES][2] = {
      {13, 0}
   ,  { 0, 2}
   ,  { 0, 2}
   ,  { 1, 2}
   ,  { 9, 1}
   ,  { 9, 0}
};

#define JUMPUPDOWN_FRAMES  9
u8 const anim_jumpUp[JUMPUPDOWN_FRAMES][2] = {
      {13, 0}
   ,  {13, 0}
   ,  {15, 8}
   ,  {15, 8}
   ,  {15, 4}
   ,  {15, 2}
   ,  {15, 0}
   ,  {13, 0}
   ,  {13, 0}
};

u8 const anim_jumpDown[JUMPUPDOWN_FRAMES][2] = {
      {15, 0}
   ,  {15, 0}
   ,  {15, 2}
   ,  {15, 4}
   ,  {15, 8}
   ,  {14, 8}
   ,  {14, 0}
   ,  {13, 0}
   ,  {13, 0}
};

#define JUMPBETWEEN_FRAMES 7
u8 const anim_jumpBetween[JUMPBETWEEN_FRAMES][2] = {
      {13, 0}
   ,  {13, 0}
   ,  {17, 1}
   ,  {15, 2}
   ,  {19, 2}
   ,  {13, 2}
   ,  {13, 0}
};

#define PUSH_FRAMES 10
u8 const anim_push[PUSH_FRAMES][2] = {
      { 9, 0}
   ,  { 9, 0}
   ,  { 9, 0}
   ,  { 9, 0}
   ,  {10, 0}
   ,  {11, 0}
   ,  {11, 0}
   ,  { 0, 0}
   ,  { 3, 0}
   ,  { 9, 0}
};

#define KICK_FRAMES 10
u8 const anim_kick[KICK_FRAMES][2] = {
      {13, 0}
   ,  {17, 0}
   ,  {16, 0}
   ,  {16, 0}
   ,  {16, 0}
   ,  {16, 0}
   ,  {16, 0}
   ,  {17, 0}
   ,  {17, 0}
   ,  {13, 0}
};

#define FALL_FRAMES 4
u8 const anim_fall[FALL_FRAMES][2] = {
      {0, 5}
   ,  {0, 6}
   ,  {1, 5}
   ,  {1, 6}   
};

u8 G_frameActual = 0;            // To manage animations
u8 G_frameTotal = 0;             // To manage animations

u8 g_initDestroySoftwall = 0;    // flag var to determine when a softwall must be destroyed.

u8 g_ptx = 0;                    // tx coordinate to apply gravity after hero falls.
u8 g_pty = 0;                    // ty coordinate to apply gravity after hero falls.


//
//    TESTKEYS
//
// {keys.up, keys.right, keys.left, keys.down, keys.swap};
u8 actionForKeyPressed() {
   u8 action=0;

   if (cpct_isAnyKeyPressed()) {
      
      if(cpct_isKeyPressed(keys.up)) action=1;  
      else if(cpct_isKeyPressed(keys.right)) action=2;  
      else if(cpct_isKeyPressed(keys.left)) action=3;  
      else if(cpct_isKeyPressed(keys.down)) action=4;  
      else if(cpct_isKeyPressed(keys.swap) && g_nHeroes>1) {
         // swap only works on release.
         waitKeyUp(keys.swap);
         action=5;
      }
      else if(cpct_isKeyPressed(keys.menu)) action=10;   // To show contextual menu

   }
   return action;
}


////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
//
//    SELECT HERO SPRITE
//
//

void flipXFrame(THero *hero) __z88dk_fastcall {
   TFrame* f;
   f = hero->frame;

   if (f->faceTo != hero->faceTo) {
      cpct_hflipSpriteM0(TILE_W, TILE_H, f->sprite);
      f->faceTo = hero->faceTo;
   }
}

void selectHeroSprite(THero *hero) __z88dk_fastcall {

   u8 frameIndex=8;
   u8 hero_skin = hero->skin;
   u8 hero_nframe = hero->nframe;

 switch(hero->state) {
      case ST_standby:        break;
      case ST_standbyLadder:  frameIndex=14; break;

      case ST_walking:        frameIndex=anim_walk[hero_nframe][0]; break;
      case ST_punching:       frameIndex=anim_punch[hero_nframe][0]; break;
      case ST_pushing:        frameIndex=anim_push[hero_nframe][0]; break;
      case ST_kicking:        frameIndex=anim_kick[hero_nframe][0]; break;
      case ST_jumpingTo:      frameIndex=anim_jumpTo[hero_nframe][0]; break;
      case ST_jumpingFrom:    frameIndex=anim_jumpFrom[hero_nframe][0]; break;
      case ST_jumpingUp:      frameIndex=anim_jumpUp[hero_nframe][0]; break;
      case ST_jumpingDown:    frameIndex=anim_jumpDown[hero_nframe][0]; break;
      case ST_jumpingBetween: frameIndex=anim_jumpBetween[hero_nframe][0]; break;

      case ST_falling:        frameIndex=anim_fall[hero_nframe][0]; break;
      case ST_landing:        frameIndex=2; break;

      case ST_sleeping:       frameIndex=7; break;
      case ST_sleepingLadder: frameIndex=12; break;
   }

   hero->frame = &g_frames[hero_skin][frameIndex];
   flipXFrame(hero);

}


////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
//
//    REDRAW HERO
//
//

// Effects management for explosions and softwall destructions.
void drawEffect(u8 tx, u8 ty, u8 *sprite) {
   u8 *pvideo;

   pvideo = cpct_getScreenPtr(CPCT_VMEM_START, (tx*TILE_W)+LEFT_OFFSET, ty*TILE_H);
   cpct_drawSpriteMaskedAlignedTable(sprite, pvideo, TILE_W, TILE_H, transparentMaskTable);
}


void drawHero(THero *hero) {
   u8 *pvideo; //, *pvmemExplosion;
   
   pvideo = cpct_getScreenPtr(CPCT_VMEM_START, hero->x, hero->y);

   // Draw explosion if is required.
   if(enemyExplosion.nframe > 0) drawEffect(enemyExplosion.tx, enemyExplosion.ty, g_explosion[enemyExplosion.nframe - 1]);

   // Draw hero.
   cpct_drawSpriteMaskedAlignedTable(hero->frame->sprite, pvideo, TILE_W, TILE_H, transparentMaskTable);

   // Draw softwall destruction if is required.   
   if(softwallDestruction.nframe > 0) drawEffect(softwallDestruction.tx, softwallDestruction.ty, g_destruction[softwallDestruction.nframe - 1]);

   // Assign tiles coordenates
   hero->tx = ((hero->x)-LEFT_OFFSET)/TILE_W;
   hero->ty = hero->y/TILE_H;
}

void restoreBg(u8 px, u8 py) {
   u8 tx, ty, auxH, auxV;
   u8 *pvideo, *pvideoH, *pvideoV;//, *pvideoHV;

   // 1. Find tiles coordenates
   tx = ((px)-LEFT_OFFSET)/TILE_W;
   ty = py/TILE_H;
   
   // 2. Test if sprite is over more than one tile in horizontal and vertical
   auxH=(px-LEFT_OFFSET)%TILE_W;
   auxV=py%TILE_H;

   // 3. Pointers to video memory precalculation
   pvideo = cpct_getScreenPtr(CPCT_VMEM_START, (tx*TILE_W)+LEFT_OFFSET, ty*TILE_H);       //*** antes la Y era (ty*TILE_H)+UP_OFFSET
   if(auxH) pvideoH = cpct_getScreenPtr(CPCT_VMEM_START, ((tx+1)*TILE_W)+LEFT_OFFSET, ty*TILE_H);  //*** antes la Y era (ty*TILE_H)+UP_OFFSET
   else if(auxV) pvideoV = cpct_getScreenPtr(CPCT_VMEM_START, (tx*TILE_W)+LEFT_OFFSET, (ty+1)*TILE_H);  //*** antes la Y era (ty*TILE_H)+UP_OFFSET


   // 4. Sincronization with Raster  
   // TRICK TO AVOID FLICKERING
   // I know where is raster drawing screen thanks to global var g_nInterrupt

   waitRaster(ty);


   // 5. Draw background to erase last hero sprite.
   cpct_drawSprite (tileset[finalLevel[tx][ty]], pvideo, TILE_W, TILE_H);

   if(auxH) cpct_drawSprite (tileset[finalLevel[tx+1][ty]], pvideoH, TILE_W, TILE_H);        // if hero is between 2 tiles in horizontal.
   else if(auxV) cpct_drawSprite (tileset[finalLevel[tx][ty+1]], pvideoV, TILE_W, TILE_H);   // if hero is between 2 tiles in vertical.

}

void redrawHero(THero *hero) __z88dk_fastcall {
   restoreBg(hero->px, hero->py);  
   hero->px = hero->x;
   hero->py = hero->y;
 
   drawHero(hero);
}


////////////////////////////////////////////////////////////////////////////////
//
//    VALID MOVE
//
//
//    BASICLEVEL VALUES
//    0: Void.
//    1: Heroes
//    2: Enemy terrestial.
//    3: Enemy aereal.
//    4: Stone.
//    5: Ladder.
//    6: Soft Wall.
//    7: Solid Wall.


void updateBasicLevel(THero *hero, u8 dir) {
   u8 heroTx=hero->tx, heroTy=hero->ty;
   i8 displacementX=0, displacementY=0;


 // Calculate tryTile value:
   switch(dir){
      case TRY_LEFT:       { displacementX=-1; break; }
      case TRY_RIGHT:      { displacementX= 1; break; }
      case TRY_UP:         { displacementY=-1; break; }
      case TRY_DOWN:       { displacementY= 1; break; }
   }

   basicLevel[heroTx][heroTy] = hero->mapValue;                              // restore map value
   hero->mapValue = basicLevel[heroTx+displacementX][heroTy+displacementY];  // save new pos map value
   basicLevel[heroTx+displacementX][heroTy+displacementY] = 1;               // assign to new pos map the hero value "1";
}


// cases to test if is a valid movement, and the action that hero must do.
// items: {Direccion, heroTile, tryTile, response/acction}

#define NCASES 19
const u8 cases[NCASES][4] = {
      { 0, 0, 0, 1}  // walk
   ,  { 0, 0, 6, 1}  // walk (must destroy SOFTWALL)
   ,  { 0, 0, 2, 2}  // punch enemy 
   ,  { 0, 0, 3, 2}  // punch enemy 
   ,  { 0, 5, 2, 2}  // punch enemy from ladder
   ,  { 0, 5, 3, 2}  // punch enemy from ladder
   ,  { 0, 0, 5, 3}  // jumpTo ladder
   ,  { 0, 0, 4, 8}  // push Stone
   ,  { 0, 5, 4, 9}  // kick Stone from ladder
   ,  { 0, 5, 0, 4}  // JumpFrom
   ,  { 0, 5, 2, 4}  // JumpFrom (must kill enemy)
   ,  { 0, 5, 3, 4}  // JumpFrom (must kill enemy)
   ,  { 0, 5, 6, 4}  // JumpFrom (must destroy SOFTWALL)
   ,  { 1, 5, 5, 5}  // JumpUp
   ,  { 1, 5, 0, 5}  // JumpUp
   ,  { 2, 0, 5, 6}  // JumpDown (de plataforma  a escalera)
   ,  { 2, 5, 5, 6}  // JumpDown
   ,  { 2, 5, 0, 6}  // JumpDown (me tiro de la escalera)
   ,  { 0, 5, 5, 7}  // JumpBetween Ladders
};


u8 validMove(THero *hero, u8 dir){

   u8 i=0, direction=0, response=NO; // by default, movement not valid and direction = 0;
   u8 heroTx=hero->tx, heroTy=hero->ty;
   u8 casesDirection, casesHeroTile, casesTryTile;

   u8 heroTile = hero->mapValue, tryTile = 7;   // heroTile is mapValue and tryTile is a SolidWall by defaul.
   

   // Calculate tryTile value:
   switch(dir){
      case TRY_LEFT:
         if(heroTx!= 0) tryTile=basicLevel[heroTx-1][heroTy];  // so, if hero is in left lateral of the level, movement not posible (tryTile remains as a SolidWall)
         break;
      case TRY_RIGHT:
         if(heroTx!=10) tryTile=basicLevel[heroTx+1][heroTy];  // so, if hero is in right lateral of the level, movement not posible (tryTile remains as a SolidWall)
         break;
      case TRY_UP:
         direction=1;
         if(heroTy!= 0) tryTile=basicLevel[heroTx][heroTy-1];  // so, if hero is in right lateral of the level, movement not posible (tryTile remains as a SolidWall)
         break;
      case TRY_DOWN:
         direction=2;
         if(heroTy!= 8) tryTile=basicLevel[heroTx][heroTy+1];   // so, if hero is in right lateral of the level, movement not posible (tryTile remains as a SolidWall)
         break;
   }


   // I go through the whole array of cases to see if I am in a valid motion case, and return in response the action that the hero must perform.
   while (i<NCASES) {

      casesDirection = cases[i][0];
      casesHeroTile  = cases[i][1];
      casesTryTile   = cases[i][2];

      if(casesDirection==direction && casesHeroTile==heroTile && casesTryTile==tryTile) {
         response = cases[i][3];
         break;
      }
      i++;
   }

      // Test if is possible to push a STONE.
      if(response==8 || response==9) {
         if(dir==TRY_LEFT  && (heroTx==1 || basicLevel[heroTx-2][heroTy]!=0) ) response=NO; // The STONE is next to level lateral SOLIDWALL.
         if(dir==TRY_RIGHT && (heroTx==9 || basicLevel[heroTx+2][heroTy]!=0) ) response=NO; // There is not space to push Stone.
      }

      // Test if hero must destroy softwall
      if(response!=NO && tryTile==6) {
         g_initDestroySoftwall=1;

      }

      return(response);
}


////////////////////////////////////////////////////////////////////////////////
//
// DESTROY SOFTWALL
//
//
//
//

void drawBackground(u8 tx, u8 ty) {
   u8 *pvideo;

   pvideo = cpct_getScreenPtr(CPCT_VMEM_START, (tx*TILE_W)+LEFT_OFFSET, ty*TILE_H);
   cpct_drawSprite (tileset[finalLevel[tx][ty]], pvideo, TILE_W, TILE_H);
}

void regenerateBackground(u8 tx, u8 ty, u8 direction) {
   u8 tile; //, *pvideo;

   // 1. REGENERATE SOFTWALL DESTROYED TILE BG
   // Calculates the new background sprite of the self tile destroyed, according contiguous tiles
   // This bg will be drawed in drawHero() due to hero is moving to this tile.

   if ( ty%2==0 ) tile=0;   // bg even                // 10=00
   else tile=3;             // bg odd                 // 20=03
   if(tx>0 && basicLevel[tx-1][ty]>5) {
      tile+=1;                                        // 12=02 | 22=05
      //if(ty>0 && basicLevel[tx-1][ty-1]<6) tile-=1;   // 11=01 | 21=04
   } else if (tx==0) tile+=1;                         // 12=02 | 22=05

   // Light version
   if(cpct_getRandom_mxor_u8()<128) tile+=6;

   finalLevel[tx][ty]=tile;


   // 2. REGENERATE THE RIGHT CONTIGUOUS TILE BG TO THE DESTROYED TILE
   // Calculates background sprite of right adyacent tile from the tile destroyed, according contiguous tiles.
   
   if(tx<10) {
      if(basicLevel[tx+1][ty]==5) finalLevel[tx+1][ty]-=1;     // Removes shadow from ladder.
      if(basicLevel[tx+1][ty]<5) 
      {
         if(finalLevel[tx+1][ty]==1 || finalLevel[tx+1][ty]==4 || finalLevel[tx+1][ty]==7 || finalLevel[tx+1][ty]==10) finalLevel[tx+1][ty]-=1;     // Removes shadow type 1 (upper shadow) from background.
         if(finalLevel[tx+1][ty]==5) finalLevel[tx+1][ty]=2;                                                                                        // Removes shadow from column.
      }
      
      if(direction==F_right) {
         if(basicLevel[tx+1][ty]==0) drawBackground(tx+1, ty);
         if(basicLevel[tx+1][ty]==1) {
            redrawHero(&hero1);
            redrawHero(&hero2);
         } 
      }
   }

   // 3. UPPER TILE FROM THE DESTROYED TILE
   // Calculates upper tile sprite to the tile destroyed, according contiguous tiles

   if(ty>0) {
      if(finalLevel[tx][ty-1]==14 || finalLevel[tx][ty-1]==17) {
         finalLevel[tx][ty-1]-=1;     // Create tile solidwall without tile bellow.
         drawBackground(tx, ty-1);
      }
      else if(finalLevel[tx][ty-1]==30 || finalLevel[tx][ty-1]==32) {
         finalLevel[tx][ty-1]+=1;     // Create tile softwall without tile bellow.
         drawBackground(tx, ty-1);
      }      
   }

   // 4. LOWER TILE FROM THE DESTROYED TILE
   // Calculates lower tile sprite to the tile destroyed.
   if(finalLevel[tx][ty+1]==32 || finalLevel[tx][ty+1]==33) {
      finalLevel[tx][ty+1]-=2;     // Create tile solidwall without tile bellow.
      drawBackground(tx, ty+1);
   }

}

// Called from walking() and jumping()
void destroySoftwall(THero* hero) {
   u8 hero_ty=hero->ty;
   u8 hero_tx=hero->tx;

   // Find the tile to destroy according to hero direction
   hero_tx+=1;
   if (hero->faceTo == F_left) hero_tx-=2;

   softwallDestruction.tx=hero_tx;
   softwallDestruction.ty=hero_ty;
   softwallDestruction.nframe = 1;

   playFX(BREAK_SOFTWALL);


   // regenerates contigous tiles due to destruction.
   regenerateBackground(hero_tx, hero_ty, hero->faceTo);

   hero->mapValue=0;          // Update mapValue with 0 (void tile) to overwrite the 6 (softwall)
   g_initDestroySoftwall=0;   // Reset global g_initDestroySoftwall to do not enter here again in the same action. 
}


////////////////////////////////////////////////////////////////////////////////
//
//    WALKING
//
//

void walking_animate(THero *hero) __z88dk_fastcall {
   if(++hero->nframe == WALK_FRAMES) hero->nframe = 0;

   // to destroy SOFTWALL
   if(g_initDestroySoftwall) destroySoftwall(hero);
}

void moveHero(THero *hero) __z88dk_fastcall { 
   i8 offset = 1;
   if(hero->faceTo == F_left) offset=-1; //change displacement direction
   hero->x+=offset;
}

void walking(THero* hero) __z88dk_fastcall {
   
   u8 keepwalking=1, direction=UNDEFINED;
   static u8 G_k;

   if(G_frameActual==6) {     // This mean hero arrives to a new tile, so keys are tested.

      // gravity - check if gravity must enter in action after hero walks to contiguous tile.
      checkGravityForHeroMoves(hero);

      G_frameActual=0;
      G_k = actionForKeyPressed();

      // Hero falling management
      // if hero is on a integer tile and tile bellow is void, go to standBy, so there I will start the hero fall
      if(basicLevel[hero->tx][hero->ty+1]==0) G_k=0;
   }
   else ++G_frameActual;

   switch(G_k) {
      case 0: case 1: case 4: 
         standby_start(hero);
         break;
     
      case 2:
         direction=TRY_RIGHT;
         break;

      case 3:
         direction=TRY_LEFT;
         break;
   }
   

   if(direction!=UNDEFINED) {
      if(G_frameActual==0) {
         hero->faceTo=direction;
         if( validMove(hero, direction) !=YES ) keepwalking=0;   // If there is an enemy or not void tile, hero stop walking and goes to stand_by.
         else updateBasicLevel(hero, direction);                 // else, hero walk to the new tile and update basicLevel.
      }
       
      if(keepwalking) {
         moveHero(hero);
         walking_animate(hero);
      } else standby_start(hero); 
   }

   hero->move = YES;
   
}

void walking_start(THero *hero, u8 faceToDir) {

   hero->nframe = 0;
   hero->state = ST_walking;
   hero->faceTo   = faceToDir;
   hero->move  = YES;

   G_frameActual=6; // init to 6 to capture keys in walking()
}


////////////////////////////////////////////////////////////////////////////////
//
//    PUNCHING
//
//

void movePunch(THero *hero) __z88dk_fastcall { 
      //i8 offset = slide_punch[hero->nframe];

      i8 offsetX = anim_punch[hero->nframe][1];

      if(hero->faceTo == F_left) offsetX*=-1; //change displacement direction
      hero->x+=offsetX;
}

void punching_animate(THero *hero) __z88dk_fastcall {
   if(++hero->nframe == PUNCH_FRAMES)
      hero->nframe = PUNCH_FRAMES-1;

   if(hero->nframe == 5) {
      playFX(EXPLOSION);
      enemyExplosion.nframe=1;  // this is the exact timig to do an explosión. The explosion will be located in enemy killed position.
   }
}

void punching(THero* hero) __z88dk_fastcall {

   if( G_frameActual > 0) {
      --G_frameActual;
      movePunch(hero);
      punching_animate(hero);
      hero->move = YES;

   } else {
      // if no enemies in this level...
      if(g_nEnemies==0) drawLevelComplete();
      else standby_start(hero);
   }

}

void killEnemyIndex(THero *hero, u8 tx, u8 ty){
   u8 i;

   // Looking for the enemy, and set his state to die.
   for(i=0;i<MAX_ENEMIES;i++) {
      if(enemyList[i].state==ST_idle && enemyList[i].tx == tx && enemyList[i].ty == ty) {
         enemyList[i].state=ST_die;
         --g_nEnemies;
         break;
      }
   }

   enemyExplosion.tx=tx;
   enemyExplosion.ty=ty;
   // I'll start the this explosion doing enemyExplosion.nframe=1 with the perfect timing inside punching_animate.

   // I need to update basicLevel to remove enemy. 
   // But now, in this position is the hero, so I need to update the mapValue of hero.
   hero->mapValue=0;

}

void punching_start(THero *hero, u8 direction) {

   i8 displacementX=1;

   playFX(JUMP_TO);
   hero->nframe = 0;
   hero->state = ST_punching;
   hero->faceTo = direction;
   hero->move  = YES;

   G_frameActual=PUNCH_FRAMES;   // init cycle of punch.

   updateBasicLevel(hero, direction);
   if(direction==F_left) displacementX=-1;
   killEnemyIndex(hero, hero->tx+displacementX, hero->ty);

}


////////////////////////////////////////////////////////////////////////////////
//
//    PUSHING STONES
//
//


void pushing_animate(THero *hero) __z88dk_fastcall {
   if(++hero->nframe == G_frameTotal) hero->nframe = G_frameTotal-1;

   // this is the exact timig to star the slide stone animation.
   if(hero->nframe == 5) {
      playFX(PUSH);
      slideStone.x=hero->x;
      slideStone.y=hero->y;
      
      if(hero->faceTo==F_left) slideStone.x-=TILE_W;
      else slideStone.x+=TILE_W;

      slideStone.px=slideStone.x;
      slideStone.py=slideStone.y;
      slideStone.direction=hero->faceTo;
      slideStone.nframe=1;
   }
                                                
}

void pushing(THero* hero) __z88dk_fastcall {

   if( G_frameActual > 0) {
      --G_frameActual;
      pushing_animate(hero);
      hero->move = YES;

   } else standby_start(hero); 

}

void pushing_start(THero *hero, u8 direction, u8 hitType) {

   hero->nframe = 0;

   if(hitType==8) {
      hero->state = ST_pushing;   
      G_frameTotal=PUSH_FRAMES;
   } else {
      hero->state = ST_kicking;   
      G_frameTotal=KICK_FRAMES;
   }

   hero->faceTo = direction;
   G_frameActual = G_frameTotal;
   hero->move  = YES;

   // Hero do not change to contiguous tile, so not need to call to updateBasicLevel(hero, direction);

   // Stone slides to contiguous tile, need to update basicLevel.
   if(direction==F_left) {
      basicLevel[hero->tx-1][hero->ty]=0;
      basicLevel[hero->tx-2][hero->ty]=4;
   } else {
      basicLevel[hero->tx+1][hero->ty]=0;
      basicLevel[hero->tx+2][hero->ty]=4;
   }

}


////////////////////////////////////////////////////////////////////////////////
//
//    JUMPING TO /FROM *** (al otro salto le llamaré JUMPING BETWEEN)
//
//

// HORIZONTAL

void moveJump(THero *hero) __z88dk_fastcall { 
   i8 offsetX=0, offsetY=0;

   switch(hero->state) {

      case ST_jumpingTo:   offsetX = anim_jumpTo[hero->nframe][1];   break;
      case ST_jumpingFrom: offsetX = anim_jumpFrom[hero->nframe][1]; break;
      case ST_jumpingUp:   offsetY = -anim_jumpUp[hero->nframe][1];  break;
      case ST_jumpingDown: offsetY = anim_jumpDown[hero->nframe][1]; break;
      case ST_jumpingBetween: 
         offsetX = anim_jumpBetween[hero->nframe][1];
         G_frameTotal=JUMPBETWEEN_FRAMES;
         break;
   }

   if(hero->faceTo == F_left) offsetX*=-1; //change displacement direction

   hero->x+=offsetX;
   hero->y+=offsetY;
}

void jumping_animate(THero *hero) __z88dk_fastcall {
   if(++hero->nframe == G_frameTotal)
      hero->nframe = G_frameTotal-1;
}


void jumping(THero* hero) __z88dk_fastcall {

   if( G_frameActual > 0) {
      --G_frameActual;
      moveJump(hero);
      jumping_animate(hero);
      hero->move = YES;

      // to destroy SOFTWALL
      if(g_initDestroySoftwall) destroySoftwall(hero);

   } else standby_start(hero); 

}

void jumping_start(THero *hero, u8 direction, u8 jumpType) {

   hero->nframe = 0;

   switch(jumpType) {
      case 3: hero->state = ST_jumpingTo;      playFX(JUMP_TO);   G_frameTotal=JUMP_FRAMES;         hero->faceTo = direction; break;
      case 4: hero->state = ST_jumpingFrom;    playFX(JUMP_FROM); G_frameTotal=JUMP_FRAMES;         hero->faceTo = direction; break;
      case 5: hero->state = ST_jumpingUp;      playFX(JUMP_TO);   G_frameTotal=JUMPUPDOWN_FRAMES;   break;
      case 6: hero->state = ST_jumpingDown;    playFX(JUMP_FROM); G_frameTotal=JUMPUPDOWN_FRAMES;   break;
      case 7: hero->state = ST_jumpingBetween; playFX(JUMP_TO);   G_frameTotal=JUMPBETWEEN_FRAMES;  hero->faceTo = direction; break;
   }
   
   G_frameActual = G_frameTotal;
   hero->move  = YES;

   updateBasicLevel(hero, direction);
}


////////////////////////////////////////////////////////////////////////////////
//
//    SLEEPING
//
//

void wakeup(THero* hero) __z88dk_fastcall {
   
   if(hero->mapValue!=5) {
      // wakeUp from normal standBy
      hero->state = ST_standby;
      hero->frame = &g_frames[hero->skin][8];
   }
   else {
      // wakeUp from ladder
      hero->state = ST_standbyLadder;
      hero->frame = &g_frames[hero->skin][14];
   }

   hero->move = YES;
}

void sleeping_start(THero *hero) __z88dk_fastcall {
   if(hero->mapValue!=5) hero->state = ST_sleeping;
   else hero->state = ST_sleepingLadder;
   hero->move  = YES;

   playFX(SWAP_HERO);
}


////////////////////////////////////////////////////////////////////////////////
//
//    FALLING HERO
//
//

void landing(THero *hero) __z88dk_fastcall {
   if( G_frameActual > 0) {
      --G_frameActual;
      hero->move = YES;

   } else standby_start(hero);

}

void moveFall(THero *hero) __z88dk_fastcall { 
      u8 offsetY = anim_fall[hero->nframe][1];
      hero->y+=offsetY;
}

void falling_animate(THero *hero) __z88dk_fastcall {
   if(++hero->nframe == FALL_FRAMES)
      hero->nframe = FALL_FRAMES-1;
}

void falling(THero *hero) __z88dk_fastcall {

   if( G_frameActual > 0) {
      --G_frameActual;
      moveFall(hero);
      falling_animate(hero);
      hero->move = YES;

   } else {

      if(basicLevel[hero->tx][hero->ty+1]==0) {
         falling_start(hero);
      }
      else {
         G_frameActual=4;
         hero->state = ST_landing;
         playFX(LANDING);
      }
   }

}

void falling_start(THero *hero) __z88dk_fastcall {
   
   hero->nframe = 0;
   hero->state = ST_falling;
   G_frameActual=FALL_FRAMES;

   hero->move  = YES;

   updateBasicLevel(hero, TRY_DOWN);
}


////////////////////////////////////////////////////////////////////////////////
//
//    STANDBY
//
//

void standby(THero *hero) {

   u8 k = actionForKeyPressed();
   u8 action, direction=UNDEFINED;

   // Hero falling management
   if(basicLevel[hero->tx][hero->ty+1]==0 && hero->state==ST_standby) {
         if(k!=10) k=0; // to do not perform any action due to keypress... *** ELIMINAR EL if(k!=10)

         // Gravity - determine where need to test gravity after hero fall.
         if(hero->ty > 0) {
            g_ptx = hero->tx;
            g_pty = hero->ty - 1;
            
         } else g_ptx = 255;   // 255 used to represent no value.

         falling_start(hero);

   }

   switch(k) {
      case 0: break; // No key was pressed...

      case 1:
         direction=TRY_UP;
         break;

      case 2:
         direction=TRY_RIGHT;
         break;

      case 3:
         direction=TRY_LEFT;
         break;

      case 4: 
         direction=TRY_DOWN;
         break;

      case 5: sleeping_start(hero);  break;

      case 10: drawContextMenu();
   }


   if(direction!=UNDEFINED)
   {
      action = validMove(hero, direction); 
      if( action==YES ) { 
         walking_start(hero, direction);
         walking(hero);
      }
      else if( action==2 ) {                       // there is an enemy
         punching_start(hero, direction);          // punch enemy
      }
      else if( action>2 && action<8 ) {            // Jump to ladder, jump from ladder, jumpUp, jumpDown or jump between ladders

         // Gravity - check if gravity must enter in action after perform jumpDown from standBy. 
         if(hero->state==ST_standby && action==6 && hero->ty > 0){
            g_pty = hero->ty - 1;
            g_ptx = hero->tx;

         } else g_ptx = 255;        // 255 used to represent no value.
         
         jumping_start(hero, direction, action);   // jumping: 3=jumpTo | 4=jumpFrom | 5=jumpUp | 6=jumpDown | 7=jumpBetween
      }
      else if( action==8 || action==9) {           // there is a stone
         pushing_start(hero, direction, action);   // pushing a stone
      }

   }
}


void standby_start(THero *hero) __z88dk_fastcall {

   // Gravity - check if gravity must enter in action.
   checkGravityForHeroMoves(hero);

   // Gravity - If g_ptx!=255 I need to test gravity above the tile from hero start to fall.
   // g_ptx and g_pty can be controlled inside standBy() from start falling and jumpDown
   if(g_ptx != 255) checkGravityForHeroFalls(g_ptx, g_pty);
   g_ptx = 255;   // reset g_ptx


   if(hero->mapValue==5) hero->state = ST_standbyLadder;
      else hero->state = ST_standby;
   hero->move  = YES;
}


////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
//
//    GO TO STATE
//
//

void goToState(THero *hero) __z88dk_fastcall {

   switch(hero->state) {
      case ST_standby:           standby(hero);    break;
      case ST_standbyLadder:     standby(hero);    break;
      case ST_walking:           walking(hero);    break;
      case ST_punching:          punching(hero);   break;
      case ST_pushing:           pushing(hero);    break;
      case ST_kicking:           pushing(hero);    break;
      case ST_jumpingTo:         jumping(hero);    break;
      case ST_jumpingFrom:       jumping(hero);    break;
      case ST_jumpingUp:         jumping(hero);    break;
      case ST_jumpingDown:       jumping(hero);    break;
      case ST_jumpingBetween:    jumping(hero);    break;

      case ST_falling:     falling(hero);    break;
      case ST_landing:     landing(hero);    break;

      case ST_wakeup:      wakeup(hero);     break;      
   }

}


////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
//
//    INITIALIZATION
//
//

void initHero(THero *hero, u8 skin, u8 tx, u8 ty) {

   //THero hero;
   hero->skin = skin; // SHAOLINKID = 0 / LITTLEINDIAN = 1
   hero->x = hero->px = (tx*TILE_W)+LEFT_OFFSET;
   hero->y = hero->py = ty*TILE_H;
   hero->tx = tx;
   hero->ty = ty;
   hero->move  = NO;
   
   if( g_nHeroes>1 && skin==SHAOLINKID  ) {
      hero->state = ST_sleep;
      hero->frame = &g_frames[hero->skin][7];
   }
   else {
      hero->state = ST_standby;
      hero->frame  = &g_frames[hero->skin][8];
   }

   hero->nframe = 0;
   hero->faceTo = F_left;
   if(tx<5) hero->faceTo = F_right;
   flipXFrame(hero);
   hero->mapValue = 0;
   
   g_ptx = 255;
   g_pty = 255;
   
   drawHero(hero);


   // Init Globals
   g_initDestroySoftwall=0;
   G_frameActual = 0;
   G_frameTotal = 0;   

}
