//-----------------------------LICENSE NOTICE------------------------------------
//   ____        _           _       _____      _                                
//  |  _ \      | |         ( )     |  __ \    | |                               
//  | |_) | __ _| |__   __ _|/ ___  | |__) |_ _| | __ _  ___ ___                 
//  |  _ < / _` | '_ \ / _` | / __| |  ___/ _` | |/ _` |/ __/ _ |                
//  | |_) | (_| | |_) | (_| | \__ \ | |  | (_| | | (_| | (_|  __/                
//  |____/ \__,_|_.__/ \__,_| |___/ |_|   \__,_|_|\__,_|\___\___|                
//
//  Code & Gfx Copyright (C) 2017 Rafael Castillo (@Azicuetano)
//  Music & Fx Copyright (C) 2017 John McKlain (@elmcklain)
//  This file is part of BABA'S PALACE: an Amstrad CPC Game made with CPCTelera
//
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
//------------------------------------------------------------------------------

#ifndef _ENTITIES_H_
#define _ENTITIES_H_

#include <types.h>
#include <cpctelera.h> 

// Hero states
extern enum {
      ST_standby  = 0
   ,  ST_standbyLadder
   ,  ST_walking
   ,  ST_sleeping
   ,  ST_sleepingLadder
   ,  ST_sleep
   ,  ST_wakeup
   ,  ST_initPunch
   ,  ST_punching
   ,  ST_initPush
   ,  ST_pushing
   ,  ST_kicking
   ,  ST_initJump
   ,  ST_jumpingTo
   ,  ST_jumpingFrom
   ,  ST_jumpingUp
   ,  ST_jumpingDown
   ,  ST_jumpingBetween
   ,  ST_falling
   ,  ST_landing
   ,  ST_numStates
} EHeroStates;


typedef struct {
   u8 faceTo;        // Direction where hero is looking at
   u8* sprite;       // Sprite
} TFrame;

typedef struct {
   u8 skin;       // can be SHAOLINKID = 0 or LITTLEINDIAN = 1;
   u8 x, y;       // Actual position.
   u8 tx, ty;     // Tile position inside grid level map.
   u8 px, py;     // Previous position.
   u8 move;
   u8 state;
   TFrame *frame;
   u8 nframe;
   u8 faceTo;
   u8 jump;       // *** NO SE SI ESTO ES NECESARIO...
   u8 mapValue;   // The basicLevel value where hero is.
} THero;

extern enum {
      ST_nocreated  = 0
   ,  ST_terrestrial
   ,  ST_aerial
} EEnemyType;

extern enum {
      ST_null  = 0
   ,  ST_fall
   ,  ST_idle
   ,  ST_die
} EEnemyState;

typedef struct {
   u8 state;      // State = last 4 bits: No_create=0, falling=1, idle=2, die=3;
   u8 type;       // Type  =  1st 4 bits: No_create=0, TERRESTRIAL=1, AERIAL=2;
   u8 x, y;       // Actual position.
   u8* pvideo;    // Pointer to videomem according to x & y. 
   u8 tx, ty;     // Tile position inside grid level map. 
   u8 px, py;     // Previous position. ***NOT IN USE*** 
   u8 iEnemy;     // index who refences to an enemy [0=Skeleton | 1=Mummy | 2=Phantom | 3=Bat]
   u8* sprite;    // Pointer to character sprite. (speed up the access to the sprite)
} TEnemy;

typedef struct {
   u8 tx, ty;        // Tile position inside grid level map.
   u8 nframe;        // nº of frame to show. 0=inactive.
} TExplosion;

typedef struct {
   u8 x, y;          // Item position.
   u8 px, py;        // Item previous position.
   u8 type;          // Number that indetifies item type (hero=1 | enemy terrestrial=2 | stone=4)
   u8 direction;     // direction: for Stones: 0=left | 1=right - for enemies: index of enemy inside enemyList[]
   u8 nframe;        // nº of frame in animation array. 0=inactive.
   u8* sprite;       // Pointer to sprite.
} TSlideItem;


//////////////////////////////////////////////////////////////////////////
// FUNCTIONS
//////////////////////////////////////////////////////////////////////////

// GENERAL

u8 actionForKeyPressed();
   
// SELECT HERO SPRITE
void flipXFrame(THero *hero) __z88dk_fastcall;
void selectHeroSprite(THero *hero) __z88dk_fastcall;

// REDRAW HERO
void drawHero(THero *hero);
void restoreBg(u8 px, u8 py);
void redrawHero(THero *hero) __z88dk_fastcall;

// VALID MOVE
void updateBasicLevel(THero *hero, u8 direction);
u8 validMove(THero *hero, u8 direction);

// DESTROY SOFTWALL
void drawBackground(u8 tx, u8 ty);
void regenerateBackground(u8 tx, u8 ty, u8 direction);
void destroySoftwall(THero* hero); // __z88dk_fastcall;

// WALKING
void walking_animate(THero *hero) __z88dk_fastcall;
void moveHero(THero *hero) __z88dk_fastcall;
void walking(THero* hero) __z88dk_fastcall;
void walking_start(THero *hero, u8 faceToDir);

// PUNCHING
void movePunch(THero *hero) __z88dk_fastcall;
void punching_animate(THero *hero) __z88dk_fastcall;
void punching(THero* hero) __z88dk_fastcall;
void killEnemyIndex(THero *hero, u8 tx, u8 ty);
void punching_start(THero *hero, u8 faceToDir);

// PUSHING STONE
void pushing_animate(THero *hero) __z88dk_fastcall;
void pushing(THero* hero) __z88dk_fastcall;
void pushing_start(THero *hero, u8 direction, u8 hitType);

// JUMPING (Platform to Ladder, Ladder to Platform)
void moveJump(THero *hero) __z88dk_fastcall;
void jumping_animate(THero *hero) __z88dk_fastcall;
void jumping(THero* hero) __z88dk_fastcall;
void jumping_start(THero *hero, u8 direction, u8 jumpType);

// SLEEPING
void wakeup(THero* hero) __z88dk_fastcall;
void sleeping_start(THero *hero) __z88dk_fastcall;

// FALLING
void moveFall(THero *hero) __z88dk_fastcall;
void falling_animate(THero *hero) __z88dk_fastcall;
void falling(THero *hero) __z88dk_fastcall;
void falling_start(THero *hero) __z88dk_fastcall;

// STANDBY
//void standby(THero *hero) __z88dk_fastcall;
void standby(THero *hero);
void standby_start(THero *hero) __z88dk_fastcall;

// GO TO STATE
void goToState(THero *hero) __z88dk_fastcall;

// INITIALIZATION
void initHero(THero *hero, u8 skin, u8 tx, u8 ty);


#endif