//-----------------------------LICENSE NOTICE------------------------------------
//   ____        _           _       _____      _                                
//  |  _ \      | |         ( )     |  __ \    | |                               
//  | |_) | __ _| |__   __ _|/ ___  | |__) |_ _| | __ _  ___ ___                 
//  |  _ < / _` | '_ \ / _` | / __| |  ___/ _` | |/ _` |/ __/ _ |                
//  | |_) | (_| | |_) | (_| | \__ \ | |  | (_| | | (_| | (_|  __/                
//  |____/ \__,_|_.__/ \__,_| |___/ |_|   \__,_|_|\__,_|\___\___|                
//
//  Code & Gfx Copyright (C) 2017 Rafael Castillo (@Azicuetano)
//  Music & Fx Copyright (C) 2017 John McKlain (@elmcklain)
//  This file is part of BABA'S PALACE: an Amstrad CPC Game made with CPCTelera
//
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
//------------------------------------------------------------------------------

#include <cpctelera.h>
#include "../entities/entities.h"

#include "menu.h"
#include "game.h"

#include "../keyboard/keyboard.h"

#include "../levels/levels.h"

#include "../globals.h"
#include "../text/text.h"
#include "../utils/uitoa.h"

#include "../sprites/atlas_background.h"
#include "../sprites/atlas_idol.h"
#include "../sprites/atlas_solidwall.h"
#include "../sprites/leftside_solidwall.h"
#include "../sprites/rightside_solidwall.h"
#include "../sprites/atlas_softwall.h"
#include "../sprites/atlas_destruction.h"
#include "../sprites/atlas_ladder.h"
#include "../sprites/stone.h"

#include "../sprites/atlas_enemies.h"
#include "../sprites/peloton-sideral.h"
#include "../sprites/hero_shaolinkid.h"
#include "../sprites/hero_littleindian.h"

#include "../sprites/char.h"

#include "../audio/music.h"




THero hero1, hero2;
TEnemy enemyList[MAX_ENEMIES];

TExplosion enemyExplosion = {0,0,0}; 				// tx, ty, nframe. 0=inactive.
TExplosion softwallDestruction = {0,0,0}; 			// tx, ty, nframe. 0=inactive.
TSlideItem slideStone = {0,0,0,0,0,0,0, gfx_stone};	// x, y, px, py, direction, nframe, *sprite.


#define FALLITEM_FRAMES 	4
u8 const anim_fallItem[FALLITEM_FRAMES] = {5, 6, 5, 6};


#define SLIDESTONE_FRAMES 	4
u8 const anim_slideStone[SLIDESTONE_FRAMES] = {3, 2, 1, 1};

#define EXPLOSION_FRAMES  	3
//u8* const g_explosion[EXPLOSION_FRAMES] = { gfx_atlas_enemies_08, gfx_atlas_enemies_09, gfx_atlas_enemies_10, gfx_atlas_enemies_11 };
u8* const g_explosion[EXPLOSION_FRAMES]   = { gfx_atlas_enemies_08,    gfx_atlas_enemies_09,    gfx_atlas_enemies_10};
u8* const g_destruction[EXPLOSION_FRAMES] = { gfx_atlas_destruction_0, gfx_atlas_destruction_1, gfx_atlas_destruction_2};


// Enemies [type][frame]
u8* const enemiesGfx[5][2] = {
  { gfx_atlas_enemies_00, gfx_atlas_enemies_01 }, // [0] Skeleton - terrestrial
  { gfx_atlas_enemies_02, gfx_atlas_enemies_03 }, // [1] Mummy    - terrestrial
  { gfx_atlas_enemies_04, gfx_atlas_enemies_05 }, // [2] Phantom  - aerial
  { gfx_atlas_enemies_06, gfx_atlas_enemies_07 }, // [3] Bat      - aerial
  { gfx_peloton_sideral_0, gfx_peloton_sideral_1 }, // *** Gesture to the game Phantis from Dinamic ***
};

// Enemies type binary distribution for 100 levels: Terrestrial=direct, aerial=invers.
// 11110110 11110111 01111101 10110011 01101001 01100110 11001010 01100110 01000101 00100101 00001000 00100100 1000
u8* const enemiesDistribution[13] = {246, 247, 125, 179, 105, 102, 202, 102, 69, 37, 8, 36, 128};


// LEVEL DEFINITION: I'll take a 11x8 compress level (33bytes) and I'll transform into a 11x9 true tile level (99 Bytes)

// Basic level with tiles types (from 0 to 7). I will store here my decompress level.
// But theres is something special in this codification.
u8 basicLevel[11][9]= { {0,0,0,0,0,0,0,0,0}, {0,0,0,0,0,0,0,0,0}, {0,0,0,0,0,0,0,0,0}, {0,0,0,0,0,0,0,0,0}, {0,0,0,0,0,0,0,0,0}, {0,0,0,0,0,0,0,0,0}, {0,0,0,0,0,0,0,0,0}, {0,0,0,0,0,0,0,0,0}, {0,0,0,0,0,0,0,0,0}, {0,0,0,0,0,0,0,0,0}, {0,0,0,0,0,0,0,0,0} };

// Index Level with index from stoneList and enemyList. 
//u8 indexLevel[11][9]

// Complet level with values from 0 to 255. I will store here my final level with real tiles index.
u8 finalLevel[11][9]= { {0,0,0,0,0,0,0,0,0}, {0,0,0,0,0,0,0,0,0}, {0,0,0,0,0,0,0,0,0}, {0,0,0,0,0,0,0,0,0}, {0,0,0,0,0,0,0,0,0}, {0,0,0,0,0,0,0,0,0}, {0,0,0,0,0,0,0,0,0}, {0,0,0,0,0,0,0,0,0}, {0,0,0,0,0,0,0,0,0}, {0,0,0,0,0,0,0,0,0}, {0,0,0,0,0,0,0,0,0} };

// Array of tiles. FinalLevel contains the index of a tile inside this array.
u8* const tileset[34] = { 

			  gfx_atlas_background_00,  gfx_atlas_background_01,  // 0  1
			  gfx_atlas_background_08,  gfx_atlas_background_02,  // 2  3
			  gfx_atlas_background_03,  gfx_atlas_background_09,  // 4  5
			  gfx_atlas_background_04,  gfx_atlas_background_05,  // 6  7
			  gfx_atlas_background_00,  gfx_atlas_background_06,  // (8 legacy)  9
			  gfx_atlas_background_07,  gfx_atlas_background_00,  // 10 (11 legacy) // legacy... these tiles has been discarted.
			  gfx_atlas_solidwall_0,    gfx_atlas_solidwall_1,    // 12 13
			  gfx_atlas_solidwall_2,    gfx_atlas_solidwall_3,    // 14 15
			  gfx_atlas_solidwall_4,    gfx_atlas_solidwall_5,    // 16 17
			  gfx_atlas_ladder_0,       gfx_atlas_ladder_1,       // 18 19
			  gfx_atlas_ladder_2,       gfx_atlas_ladder_3,       // 20 21
			  gfx_atlas_ladder_4,       gfx_atlas_ladder_5,       // 22 23
			  gfx_atlas_idol_0,         gfx_atlas_idol_1,         // 24 25
			  gfx_atlas_idol_2,         gfx_atlas_idol_3,         // 26 27
			  gfx_stone,              	gfx_stone,                // 28 29 // legacy. 29 is a stone due to delete lantern.
			  gfx_atlas_softwall_0,     gfx_atlas_softwall_1,     // 30 31
			  gfx_atlas_softwall_2,     gfx_atlas_softwall_3,     // 32 33
			};


u8 g_nHeroes = 0;		// Number of heroes in a level.
u8 g_nEnemies = 0;		// Number of enemies in a level.
u8 g_finishGame = 0;	// Manage when the game is finished and return to menu.
u8 g_nInterrupt = 0;	// Manage Interrupt and locate raytrace


// Declarations to prevent warnings
void cpct_scanKeyboard_if();
u16 cpct_removeInterruptHandler();


//
//    INTERRUPTS
//

void gameInterruptHandler() {

	//
	// MUSIC MANAGEMENT (Awesome Music By McKlain!!!)
	// 
   	if (++g_nInterrupt == 6) {
    	cpct_akp_musicPlay();
    	cpct_scanKeyboard_if();
    	g_nInterrupt = 0;
   	}
}



void gameInit() {

  	cpct_waitVSYNC ();
  	cpct_clearScreen(0xc0c0);

	// Manage when the game is finished and return to menu
	g_finishGame = NO;

}


u8 getLevelCompressPiece(u8* level, u8 posX, u8 posY) {

  u16 index;
  u8 value=0, bit2=0, bit1=0, bit0=0, indexTrio;

  indexTrio = (LEVEL_COMPRESS_W * posY ) + posX;
  index=indexTrio*3;

  bit2 = cpct_getBit (level, index);
  bit1 = cpct_getBit (level, index+1);
  bit0 = cpct_getBit (level, index+2);

  if(bit2>0) value+=4;
  if(bit1>0) value+=2;
  if(bit0>0) value+=1;

  return (value);
}
  	
// 2. Draw FINAL LEVEL usign BASIC LEVEL.
// The final conains the exact tiles to draw a true eyecandy level.

// This time, I need iterate from top to bottom 
// in order to create the enemies in the correct order
// and avoid VSync flickering.

// complete==1 (true)    The whole level is generated with procedural techniques.
// complete==0 (false)   Nothing is generated, but a small area of level is drawn to erase the info window CONTEXTUAL MENU.
// * Unifying the level generation and the deletion of the info window, 159 bytes are saved)

void drawLevel(u8 complete) {

	u8 seed = g_nLevel;
	u8 i=0, n, tile=0, ntile, x, y, xScreen, yScreen, xIdol=0, yIdol=0;
	u8 total = (LEVEL_TILES_H)*(LEVEL_TILES_W);
	u8 terrestrialEnemy=1, aerialEnemy=1;

	u8 *pvideo;


	// -- ENEMIES GFX SELECTION 
	//
	// Enemy GFX depends from level number and the enemiesDistribution Array.
	// 0=Mummy | 1=Skeleton 
	if( cpct_getBit (enemiesDistribution, g_nLevel-1 ) ==0 ) terrestrialEnemy = 0;
	// 2=Phantom | 3=Bat
	if( cpct_getBit (enemiesDistribution, 99-(g_nLevel-1) ) ==0 ) aerialEnemy = 0;
	aerialEnemy+=2;
	

	//-- LEVEL GENERATION
	//
	if(!complete) i=35;	// No draw a complet level, only just to clear context menu.
	
	cpct_setSeed_lcg_u8(seed);
	
  	do {
		x = i % LEVEL_TILES_W;
		y = i / LEVEL_TILES_W;
	  
		xScreen = (x*TILE_W)+LEFT_OFFSET;
		yScreen = y*TILE_H;

		ntile=basicLevel[x][y];


		if(complete) {

			// PROCEDURAL LEVEL GENERATION
		  	// Checking contiguous tiles to generate the final tile.

		  	//Looking for an idol in this level, and then init xIdol and yIdol.
		  	for(n=0;n<ARRAY_IDOLS_SIZE;n+=2) {
		  		if(idols[n]==g_nLevel-1) {
		  			xIdol=(idols[n+1]>>4);
		  			yIdol=(idols[n+1] & 0b00001111);
		  		}
		  	}


			// BACKGROUND
			if(ntile<5) {

				// If this tile has backgroung.
				if ( y%2==0 ) tile=0;  // bg even   	                                    // 10=00
				else tile=3;           // bg odd	    			                        // 20=03
				if(x>0 && basicLevel[x-1][y]>5) {
					tile+=1;																// 12=02 | 22=05
				} else if (x==0) tile+=1;                                                   // 12=02 | 22=05

				// Light version
				if(cpct_getRandom_lcg_u8(0)<128) tile+=6;                                   // 10->13=00->06 ; 11->14=01->07; 12->15=02->08
																							// 20->23=03->09 ; 21->24=04->10; 22->25;05->11

				// Beautify with columns
				if(x>=2 && x<=9 && (columns[g_nLevel-1] & 0x01<<(9-x)) ) {
					tile=2;								// normal column
					if(basicLevel[x-1][y]>5) tile=5;	// shadow column
				}

				// Beautify with idol
				if(xIdol!=0) {
					if( x==xIdol && y==yIdol) tile=27;
					if( x==xIdol-1 && y==yIdol) tile=26;
					if( x==xIdol && y==yIdol-1) tile=25;
					if( x==xIdol-1 && y==yIdol-1) tile=24;
				}

			}

		  	// LADDER (18, 19, 20, 21, 22, 23)
		 	else if(ntile==5) {
				tile=18;																	// 18
				if(y>0 && basicLevel[x][y-1]==5 && basicLevel[x][y+1]==5) tile=20; 	    	// 20
			  	else if(y>0 && basicLevel[x][y-1]==5) tile=22;                              // 22
				if (x==0 || (x>0 && basicLevel[x-1][y]>5) )  tile+=1;  		                // 19 | 21 | 23
		  	}

		  	// SOFTWALL (30, 31, 32, 33)
		  	else if(ntile==6) {
		  		tile=32;																	// 32
				if(y>0 && basicLevel[x][y-1]<6) tile=30;                                    // 30

				// If up there is a solidwall, the softwall tile is special:
				if( y<8 && basicLevel[x][y+1]<5) tile+=1;                                   // 31 || 33 <- antes <6

		  	}

		  	// SOLID WALL (12, 13, 14, 15, 16, 17)
		  	else if(ntile==7) {
		  		tile=15;																	// 15
				if(y>0 && basicLevel[x][y-1]!=7) tile=12;                                   // 12

				// I assume that a supposed row 10 would consist in solidwall
				if( y<8 && basicLevel[x][y+1]<5) tile+=1;                                   // 13 | 16 <- antes <6
				else if (y<8 && basicLevel[x][y+1]==6) tile+=2;                             // 14 | 17

		  	}


		  	// Before draw first left tile, I draw the leftside solidwall. (2x22px) No RIGHT_OFFSET offset neeeded.
			if(x==0){
				pvideo = cpct_getScreenPtr(CPCT_VMEM_START, x*TILE_W, yScreen);
				cpct_drawSprite (gfx_leftside_solidwall, pvideo, LEFT_OFFSET, TILE_H);
			}

		}


		

		///////////////
		// DRAW TILE
		///////////////
		if(!complete) tile = finalLevel[x][y];	// if we are drawing tiles over context menu:
		pvideo = cpct_getScreenPtr(CPCT_VMEM_START, xScreen, yScreen);
		cpct_drawSprite (tileset[tile], pvideo, TILE_W, TILE_H);

		finalLevel[x][y]=tile;					// Store in finalLevel the index who references a tile in tileset



		// BG are drawn. Now I test if I've one objetcs over it.

		// 1. STONES
		if(ntile==4) {
			cpct_drawSpriteMaskedAlignedTable(gfx_stone, pvideo, TILE_W, TILE_H, transparentMaskTable);           
		}

		// 2. HEROES & ENEMIES
		// - Only when a complete level is generated.
		if(complete && ntile<4) {
			if (ntile==1) { // Main Characters (from 1 to 2 in a level).
				//*** ATENTION, this init system will fail in a level with a hero in x=0, y=0.
				if(hero1.tx==0 && hero1.ty==0) { // if no hero was defined...
					// init coordinates Shaoling.
					hero1.tx=x;
					hero1.ty=y;
					++g_nHeroes;
				} else {
					// init coordinates Indian.
					hero2.tx=x;
					hero2.ty=y;
					++g_nHeroes;
				}
			}


			// ENEMIES INIT MANAGEMENT
			if ((ntile==2 || ntile==3) && g_nEnemies<MAX_ENEMIES-1){

				// On this tile there is an enemy.
				enemyList[g_nEnemies].state  = ST_idle;      				// init enemy.
				if(ntile==2) enemyList[g_nEnemies].type = ST_terrestrial;  	// terretrial enemy.
				else         enemyList[g_nEnemies].type = ST_aerial;  		// aerial enemy.				
				enemyList[g_nEnemies].x  = xScreen;      			  		// Coordinates position (Screen Coordinates).
				enemyList[g_nEnemies].y  = yScreen;
				enemyList[g_nEnemies].pvideo = pvideo;
				enemyList[g_nEnemies].tx = x;                   // Tile position inside grid level map.
				enemyList[g_nEnemies].ty = y;

				if(ntile==2) enemyList[g_nEnemies].iEnemy = terrestrialEnemy;
				else 		 enemyList[g_nEnemies].iEnemy = aerialEnemy;

				// *** Gesture to the game Phantis from Dinamic ***
				if(g_nLevel==1) enemyList[g_nEnemies].iEnemy = 4; // sidereal PELOTRONES

				enemyList[g_nEnemies].sprite=enemiesGfx[enemyList[g_nEnemies].iEnemy][1]; // pointer to sprite asigned.
				cpct_drawSpriteMaskedAlignedTable(enemyList[g_nEnemies].sprite, pvideo, TILE_W, TILE_H, transparentMaskTable);

				++g_nEnemies;
			}
		}


		// After draw last right tile, I draw the rightside solidwall (4x22px)
		if(complete && x==LEVEL_TILES_W-1){
			pvideo = cpct_getScreenPtr(CPCT_VMEM_START, ((x+1)*TILE_W)+LEFT_OFFSET, y*TILE_H);
			cpct_drawSprite (gfx_rightside_solidwall, pvideo, 2, TILE_H);
		}


		i++;

		// Manage var i to only draw the level over contextual menu:
		if(!complete) {
			if(i==42) i=46;
			else if(i==53) i=57;
			else if(i==64) i=total;
		}

	} while (i<total);

}

void createLevel(u8 nLevel) {

	u8 ntile, x, y, i=0, tile=0;
	u16 pLevel;
	char txtLevel[] = "LEVEL 100";
	char ctile[6];  // 5 are the maximus digits suported by uitoa().

	// compressLevel is an Array starting on 0, but nLevel start on 1. So...
	--nLevel;

	// reset number of heroes & enemies.
	g_nHeroes=0;
	g_nEnemies=0;



	// 1. Generate basic level. Read the compress level and generate a basic level with types of tiles.
	// 2. Generate final level. Read the basic level to generate the final level using procedural tecnics.

	// 1. GENERATE THE BASIC LEVEL.
	// This basic level is a exact copy of my compress level, but inside a u8 array.
	// This makes quickliy test type of tiles to generate the final procedural level
	// and, later, consult any type o tile.

	i = (LEVEL_TILES_H)*(LEVEL_TILES_W)-1;
	do {
		x = i % LEVEL_TILES_W;
		y = i / LEVEL_TILES_W;

		if( y == 8 ) basicLevel[x][y]=7; 	// The last row is allways 'SolidWall'.
	  	else {

			// ntile stores original tile content.
			// each level has a size of 33 bytes, so I get de pointer to first level and add the level number * 33.
			pLevel=nLevel*33;
			ntile = getLevelCompressPiece(compressLevels+pLevel, x, y);
			uitoa(ntile, ctile);
			basicLevel[x][y]=ntile;
	  }

	} while ( i--);



	drawLevel(1); // 1 to draw all level / 0 to draw only over contextual menu.
	

	//
	// Draw level text
	//	

	ntile = ++nLevel;   // Levels on HUD are from 1 to 100.
	uitoa(ntile, ctile);
	
	if (nLevel<10) {
		txtLevel[6]='0'; // '0'
		txtLevel[7]=ctile[0]; //ctile[0];
		txtLevel[8]=' '; // ' '
	} else if (nLevel<100) {
		txtLevel[6]=ctile[0];
		txtLevel[7]=ctile[1];
		txtLevel[8]=' ';
	}

	drawText(txtLevel, 2, 187, COLORTXT_WHITE, NORMALHEIGHT, TRANSPARENT);
}





void initEnemies() {
	u8 i;

	for(i=0;i<MAX_ENEMIES;i++) enemyList[i].state = ST_null; // Inicialize to noinit state.
	
}

////////////////////////////////////////////////////////////////////////////////
//
//    CONTEXT MENU
//
//

void drawContextMenuWindow(){
    
    cpct_waitVSYNC ();

    // -- WHITELINES: precalculated pattern color (15,15) = 0xff
    cpct_drawSolidBox ((u16*)62099, 0xff, 42-2, 2);
    cpct_drawSolidBox ((u16*)62659, 0xff, 42-2, 2);
    
    // -- BLACKLINES: precalculated pattern color (1,1) = 0xc0
    cpct_drawSolidBox ((u16*)49891, 0xc0, 42-2, 2);
    cpct_drawSolidBox ((u16*)58563, 0xc0, 42-2, 2);

    // -- BLUEBOX: precalculated pattern color (3,3) = 0xcc
    cpct_drawSolidBox ((u16*)53987, 0xcc, 42-2, 58-8);

    // -- LEFT & RIGHT BORDERS: precalculated pattern (15, 1) = 0xea, (15,1) = 0xd5
    // left vertical line
    cpct_drawSolidBox ((u16*)53986, 0xea, 1, 58-8);
    
    // top right corner
    cpct_drawSolidBox ((u16*)49931, 0xea, 1, 2);
    
    // bottom right corner
    cpct_drawSolidBox ((u16*)58603, 0xea, 1, 2);

    // right vertical line
    cpct_drawSolidBox ((u16*)54027, 0xd5, 1, 58-8);

    // top left corner
    cpct_drawSolidBox ((u16*)49890, 0xd5, 1, 2);

    //bottom left corner
    cpct_drawSolidBox ((u16*)58562, 0xd5, 1, 2);
}

void drawLevelComplete() {
 	cpct_keyID key;
 	u8 optionSelected = NO;

	// Music Level Clear
    cpct_akp_musicInit(G_music_fanfarria);

	drawContextMenuWindow();

	drawText("LEVEL CLEAR", 28, 78,  COLORTXT_ORANGE, DOUBLEHEIGHT, OPAQUE);
	g_nLevel++;
    drawCodeLevel(20, 98);
    drawText("Press ACTION...", 26, 110,  COLORTXT_WHITE, NORMALHEIGHT, OPAQUE);

	do {
        key = waitAKey();
        if(key==keys.swap) playFX(CLICK);
        waitKeyUp(key);
        if(key==keys.swap) {
        	// playFX(SELECT);
        	optionSelected=YES;
        }

    } while (optionSelected==NO);
    g_finishGame=NEXTLEVEL;

    
    if(g_nLevel>100) {
    	// Launch the ending of Baba's Palace!
		drawEnding();

    	// Can't be over 100 level...
    	g_nLevel--;
    	// Go to main menu.
    	g_finishGame=MAINMENU;
    }
}

void drawContextMenu() {

    cpct_keyID key;
    u8 optionSelected = NO;

	playFX(OPEN_MENU);

    drawContextMenuWindow();
    drawCodeLevel(20, 78);
    drawText("1. Continue", 28, 92,  COLORTXT_WHITE, NORMALHEIGHT, OPAQUE);
    drawText("2. Restart",  28, 102, COLORTXT_WHITE, NORMALHEIGHT, OPAQUE);
    drawText("0. Quit",     28, 112, COLORTXT_WHITE, NORMALHEIGHT, OPAQUE);

    do {
        key = waitAKey();
        if(key==Key_1 || key==Key_Esc || key==Key_2 || key==Key_0) playFX(SELECT);
        waitKeyUp(key);

        if(key==Key_1 || key==Key_Esc) {

            drawLevel(0); // 1 to draw all level | 0 to draw only over contextual menu area.

			redrawHero(&hero1);
            if(g_nHeroes>1) redrawHero(&hero2);
            optionSelected=YES;            

        }
        if(key==Key_2) {
            g_finishGame=RESTART;
            optionSelected=YES;
        }
        if(key==Key_0) {
            g_finishGame=MAINMENU;
            optionSelected=YES;
        }

    } while (optionSelected==NO);
}

// I'm working syncronized with raster. 
// This function waits until raster pass the area where I want to draw a sprite.
// Much more faster than waitVSync
void waitRaster(u8 ty) {
	//*** SYNK WITH RASTER
	if(ty<3) while(g_nInterrupt<4);
	else {
		if(ty>=3 && ty<=5) while(g_nInterrupt<5);
		else while(g_nInterrupt>0);
	}
}


////////////////////////////////////////////////////////////////////////////////
//
//    GRAVITY SYSTEM
//
//

void drawSliderStone() {
   u8 *pvideo, xOffset=anim_slideStone[slideStone.nframe - 1];
   

   // draw bg
   restoreBg(slideStone.px, slideStone.py);

   // slide (if direction == F_left == 0)
   if(slideStone.direction==0) slideStone.x-=xOffset;
   else slideStone.x+=xOffset;
   slideStone.px=slideStone.x;
   slideStone.py=slideStone.y;

   // draw stone
   pvideo = cpct_getScreenPtr(CPCT_VMEM_START, slideStone.x, slideStone.y);
   cpct_drawSpriteMaskedAlignedTable(gfx_stone, pvideo, TILE_W, TILE_H, transparentMaskTable);
}

void drawItem(TSlideItem *item) {
   
   u8 *pvideo;
   
   waitRaster(item->y/TILE_H);
   // draw bg
   restoreBg(item->px, item->py);

   // draw stone
   pvideo = cpct_getScreenPtr(CPCT_VMEM_START, item->x, item->y);
   cpct_drawSpriteMaskedAlignedTable(item->sprite, pvideo, TILE_W, TILE_H, transparentMaskTable);
}

void ItemFall(TSlideItem *item) {
	
	item->nframe=0;
	while (item->nframe < FALLITEM_FRAMES) {
		
		item->px = item->x;
		item->py = item->y;

		item->y += anim_fallItem[item->nframe];
		drawItem(item);
		item->nframe++;
	}
	item->nframe=0;

	if(gravityItem(item) == NO) playFX(LANDING);

}

// returns 1 if gravity is applied, 0 if not.
u8 gravityItem(TSlideItem *item){
	u8 tx, ty, result=0;

	tx = ((item->x)-LEFT_OFFSET)/TILE_W;
   	ty = item->y/TILE_H;

	if(basicLevel[tx][ty+1]==0) {

		//update basicLevel
		basicLevel[tx][ty+1]=basicLevel[tx][ty];
		basicLevel[tx][ty]=0;


		// if item is a sleep hero: His internal y, ty and py needs to be updated.
		if(item->type==1) {
			// if hero1 is sleeping...
			if(hero1.state==5) {
				hero1.y  = (ty+1)*TILE_H;
				hero1.ty =  ty+1;
				hero1.py = hero1.y;
			} else {
				hero2.y  = (ty+1)*TILE_H;
				hero2.ty =  ty+1;
				hero2.py = hero2.y;
			}
		}

		// if item is a terrestrial enemy: His internal y, ty and pvideo needs to be updated.
		if(item->type==2) {
			enemyList[item->direction].y  = (ty+1)*TILE_H;
			enemyList[item->direction].ty = ty+1;
			enemyList[item->direction].pvideo = cpct_getScreenPtr(CPCT_VMEM_START, enemyList[item->direction].x, enemyList[item->direction].y);
		}

		result=1;
		ItemFall(item);

	}

	return(result);
}

void doGravityAbove(u8 tx, u8 ty) {

	u8 i;
	TSlideItem fallingItem;

	do{

		// init falling item.
		fallingItem.px = fallingItem.x = tx*TILE_W + LEFT_OFFSET;
		fallingItem.py = fallingItem.y = ty*TILE_H;
		fallingItem.nframe = 0;
		fallingItem.type=basicLevel[tx][ty];

		switch(fallingItem.type){


			// HERO SLEEPED
	   		case 1:	   			
	      		// if hero==ST_sleep && != ladder, hero must fall.
	      		if(hero1.state==5 && hero1.mapValue!=5) {
	      			fallingItem.sprite = gfx_hero_shaolinkid_07;
	      		}
	      		else if(hero2.state==5 && hero2.mapValue!=5) { 
	      			fallingItem.sprite = gfx_hero_littleindian_07;
	      		} else ty=255;	// not need to continue testing tiles.
	         	break;

			// TERRESTRIAL ENEMY
			case 2:
				// identifying enemy
				for(i=0;i<MAX_ENEMIES;i++) {
	      			if(enemyList[i].state==ST_idle && enemyList[i].tx == tx && enemyList[i].ty == ty) {
	      				fallingItem.direction = i;
	      				fallingItem.sprite = enemyList[i].sprite;
	         			break;
	      			}
	   			}
	   			break;

	   		// STONE
	   		case 4:
	      		fallingItem.sprite = gfx_stone;
	         	break;

	   		default:
	   			ty=255;	// not need to continue testing tiles.
		}

		if(ty==255) break;					// exit do..while
		else gravityItem(&fallingItem);		// apply gravity

		--ty;
	} while (1);
}


// Gravity management when a stone has been pushed
void checkGravityForFallingStone() {
	u8 tx, ty, ptx, pty;

	tx = ((slideStone.x)-LEFT_OFFSET)/TILE_W;
   	ty = slideStone.y/TILE_H;


   	//
	// CHECKING IF ITEMS ABOVE STONE MUST FALL
	//

	// Find where was the stone previous slide action (direction == F_left == 0)
	if(slideStone.direction==0) ptx=tx+1;
	else ptx=tx-1;
	pty=ty-1;

	// if above the Stone there is an item that must fall...
	if(basicLevel[ptx][pty]==1 || basicLevel[ptx][pty]==2 || basicLevel[ptx][pty]==4) doGravityAbove(ptx, pty);


	//
	// CHECKING IF PUSHED STONE MUST FALL
	//

	gravityItem(&slideStone);

}

// Gravity management when hero moves from one tile to other.
void checkGravityForHeroMoves(THero *hero){
	
	// Determine the tile that I must check to see if I have to trigger a drop of objects.
	u8 ptx=hero->tx, pty=hero->ty;
	
	if(pty>0) {
		--pty;	// for check uppers tiles.

		// if direction == 0 (== left) i need to check the upper right tile. Else the upper left tile.
		if(hero->faceTo==0) ++ptx;
		else --ptx;

		// if there is an item that must fall...
		if(basicLevel[ptx][pty]==1 || basicLevel[ptx][pty]==2 || basicLevel[ptx][pty]==4) doGravityAbove(ptx, pty);
	}

}

// Gravity management when hero falls.
void checkGravityForHeroFalls(u8 ptx, u8 pty){
	// if there is an item that must fall...
	if(basicLevel[ptx][pty]==1 || basicLevel[ptx][pty]==2 || basicLevel[ptx][pty]==4) doGravityAbove(ptx, pty);
}

////////////////////////////////////////////////////////////////////////////////
//
//    MAIN GAME LOOP
//
//

void gameLoop() {

	u8 i=0, j=0, moved=0, nextFrame=0, frameloop=0, indexEnemy=0, enemiesToMove=MAX_ENEMIES_MOVING, movingEnemies=1;
	u8 bgtile, k=1, lastEnemy=0, *tile[MAX_ENEMIES_MOVING], *sprite[MAX_ENEMIES_MOVING], enemyListIndex[MAX_ENEMIES_MOVING];
	u8 *pvideo;

	
	u8 yEnemy[MAX_ENEMIES_MOVING];// *** RASTER


	// pre-init heroes
	hero1.tx=0;
	hero1.ty=0;
	hero2.tx=0;
	hero2.ty=0;
	initEnemies();

	// init slideStone
	slideStone.sprite = gfx_stone;


	// GENERATE LEVEL: calculate initial Hero tile coordinates and inicializes g_nHeroes (number of heroes in this level)..
	createLevel(g_nLevel);
	fade(IN);


	// Music inGame
  	cpct_akp_musicInit(G_music_ingame);
  	// To get sincro with raster.
    cpct_waitVSYNC();
    __asm
                halt
                halt
    __endasm;
    cpct_waitVSYNC();
    g_nInterrupt = 0;

  	cpct_setInterruptHandler (gameInterruptHandler);



	initHero(&hero1, SHAOLINKID, hero1.tx, hero1.ty);
	if(g_nHeroes>1) initHero(&hero2, LITTLEINDIAN, hero2.tx, hero2.ty);


	while (g_finishGame==NO) {
	

		//
		// ENEMY EXPLOSION
		//
		if(enemyExplosion.nframe > 0) {
			if(++enemyExplosion.nframe == EXPLOSION_FRAMES+1) enemyExplosion.nframe = 0;			// explosion ends.
		}
		//
		// SOFTWALL DESTRUCTION
		//
		if(softwallDestruction.nframe > 0) {

			if(softwallDestruction.nframe==1) {
				// Check if this tile must be beautify with columns
	   			if(softwallDestruction.tx>=2 && softwallDestruction.tx<=9 && (columns[g_nLevel-1] & 0x01<<(9-softwallDestruction.tx)) ) {
		      		bgtile=2;                             											// normal column
	      			if(basicLevel[softwallDestruction.tx-1][softwallDestruction.ty]>5) bgtile=5;  	// shadow column

	      			finalLevel[softwallDestruction.tx][softwallDestruction.ty]=bgtile;
	   			}
			}
			if(++softwallDestruction.nframe == EXPLOSION_FRAMES+1) softwallDestruction.nframe = 0;	// destruction ends.
		}
		//
		// SLIDE STONE
		//
		if(slideStone.nframe > 0) {
			drawSliderStone();
			if(++slideStone.nframe == SLIDESTONE_FRAMES+1) {
				slideStone.nframe = 0;	// slide ends.
				checkGravityForFallingStone();
			}
		}

   		//
   		// HERO ANIMATION
   		// 

		moved=0;

		goToState(&hero1);
		if (hero1.move) { // if hero1->move!=0 = hero1->move!=NO
			selectHeroSprite(&hero1);
			redrawHero(&hero1);
			hero1.move=NO;
			++moved;
		}

		// If this level has 2 heroes:
		if(g_nHeroes>1) {

			goToState(&hero2);
			if (hero2.move) {
				selectHeroSprite(&hero2);
				redrawHero(&hero2);
				hero2.move=NO;
				++moved;
		   }
		   
		  // Manage the swap between heroes:
		   if(hero1.state==ST_sleeping || hero1.state==ST_sleepingLadder) {
				hero1.state=ST_sleep;
				hero2.state=ST_wakeup;
				++moved;
		   }
		   else if(hero2.state==ST_sleeping || hero2.state==ST_sleepingLadder) {
				hero2.state=ST_sleep;
				hero1.state=ST_wakeup;
				++moved;
		   }
		}

		// delay when no hero has moved, to not acceletare the idle animation of enemies.
		if(moved==0) { 
				__asm
				halt
				halt
				halt
				halt
				halt
				halt
				__endasm;
		}


		//
		// ENEMIES IDLE ANIMATON
		//

	  	if(movingEnemies && g_nEnemies>0){

			lastEnemy=MAX_ENEMIES_MOVING;

			// 1. Precalculate info
			// 1st time: indexEnemy=0, i=0, lastEnemy=0;

			while (indexEnemy < MAX_ENEMIES_MOVING) {
			  	// If enemy state == idle: save for later the sprite, pvideo, tile bg and yEnemy.
				
				if(enemyList[i].state == ST_idle) {
						enemyListIndex[indexEnemy] = i;
						sprite[indexEnemy] = enemiesGfx[enemyList[i].iEnemy][nextFrame];
						tile[indexEnemy] = tileset[finalLevel[enemyList[i].tx][enemyList[i].ty]];
						yEnemy[indexEnemy] = enemyList[i].ty; // *** RASTER
						
						++indexEnemy;
				} 
			  	++i;

			  	// if try to move all enemies, exit from while.
			  	if(i == MAX_ENEMIES-1) {
					lastEnemy = MAX_ENEMIES_MOVING - indexEnemy;
					movingEnemies=0;
					break;
			  	}
			}



			for(k=0;k<lastEnemy;k++){

				waitRaster(yEnemy[k]);

				pvideo = cpct_getScreenPtr(CPCT_VMEM_START, enemyList[enemyListIndex[k]].x, enemyList[enemyListIndex[k]].y );
				cpct_drawSprite (tile[k], pvideo, TILE_W, TILE_H);
				cpct_drawSpriteMaskedAlignedTable(sprite[k], pvideo, TILE_W, TILE_H, transparentMaskTable);
			}

			// restart index.
			indexEnemy=0;
		
		
	  	} else {					// delay when no enemy is moving to get a constant framerate.
			__asm
			halt
			halt
			halt
			halt
			halt
			halt

			halt
			halt
			halt

			__endasm;
	  	}

	  	++frameloop;
	  	if(frameloop==12) {			// every 12 loops, enemies changes his idle gfx
			i=0;
			frameloop=0;
			j++;
			nextFrame=j%2;
			movingEnemies=1;
	  	}
	}

	// FINISH GAME

	cpct_removeInterruptHandler(); 	// Removes Interrupt Handle (no more its needed).
    cpct_akp_stop();               	// Stop music.

	return;

}

void game(){

	do {
		gameInit();
		gameLoop();
		fade(OUT);

	} while (g_finishGame!=MAINMENU);
	

}