//-----------------------------LICENSE NOTICE------------------------------------
//   ____        _           _       _____      _                                
//  |  _ \      | |         ( )     |  __ \    | |                               
//  | |_) | __ _| |__   __ _|/ ___  | |__) |_ _| | __ _  ___ ___                 
//  |  _ < / _` | '_ \ / _` | / __| |  ___/ _` | |/ _` |/ __/ _ |                
//  | |_) | (_| | |_) | (_| | \__ \ | |  | (_| | | (_| | (_|  __/                
//  |____/ \__,_|_.__/ \__,_| |___/ |_|   \__,_|_|\__,_|\___\___|                
//
//  Code & Gfx Copyright (C) 2017 Rafael Castillo (@Azicuetano)
//  Music & Fx Copyright (C) 2017 John McKlain (@elmcklain)
//  This file is part of BABA'S PALACE: an Amstrad CPC Game made with CPCTelera
//
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
//------------------------------------------------------------------------------

#include "text.h"
#include "../globals.h"
#include "../sprites/char.h"
#include "../utils/uitoa.h"


 //Font Sprite Size
#define FONT_W 2
#define FONT_H 9

u8 const swapColors[6][4] = {
        {0x55, 0xea, 0xd5, 0xff},   // Bright White
        {0x15, 0x6a, 0x95, 0x3f},   // Bright Yellow
        {0x11, 0x62, 0x91, 0x33},   // Orange
        {0x45, 0xca, 0xc5, 0xcf},   // Sky Blue
        {0x54, 0x38, 0xd4, 0xfc},   // Bright Red
        {0x14, 0x68, 0x94, 0x3c}    // Mauve
};


// Transparent mask table.
cpctm_createTransparentMaskTable(transparentMaskTable, 0x0100, M0, 0);



// Function from CPCtelera examples > medium > textScroll
// Copyright (C) 2015 ronaldo / Fremos / Cheesetea / ByteRealms (@FranGallegoBR)
// Created in collaboration with Roald Strauss (aka mr_lou / CPCWiki)
//
// Waits n times for a VSYNC signal (1/50s). After each VSYNC signal,
// if waits for 2 interrupts (using HALT 2 times) to ensure that VSYNC
// stops being active until looking for the next VSYNC signal.

void wait_n_VSYNCs(u8 n) {
    do {
        // Wait for 1 VSYNC signal
        cpct_waitVSYNC();

        // If we still have to wait for more VSYNC signals, wait first
        // for this present VSYNC signal to become inactive. For that,
        // wait for 2 system interrupts using ASM halt instruction.
        if (--n) {
            __asm__("halt");
            __asm__("halt");
        }
    } while(n);
}



//
// PRINT SPRITE TEXT
//
// I've created a special character set with numbres, lowcases, uppercases and some puntuation symbols.
// All my charset are like Amstrad char set, but with some exceptions.
// 1) My 1sr char (char[0]) is Amstrad char[48], so I've to work with an offset of 48.
// 2) My charset has 75 chars.
// 3) Almost all chars are 4px wide, except 3 especial narrow who are 2px wide: ' i l
// 4) Some chars has not coincidence with Amstrad index because they are out of my 75 chars range. 
// These exceptions are: . , " ! ( ) & / '
// 
// color: to print chars in a special predefined color.
// size: To print normal chars (1) or double height chars (2).


//void drawText(u8 text[], u8 xPos, u8 yPos, u8 color, u8 size) {
void drawText(const u8 text[], u8 xPos, u8 yPos, u8 color, u8 size, u8 transparent) {

    u8* pvideo;
    u8* pChar;
    u8 character;
    u8 x = 0; // general purpose.
    u8 i = 0; // chage color bucle.
    u8 color1=0, color2=0, color3=0, color4=0; // Store pair of pixel for print chars in a predefined color.
    u8 colorchar[2 * 9 * 2]; // Store a copy of a char in a specific color. // need double memory for double heigth chars.
    u8 pos; // for precalculated values.

    color1 = swapColors[color][0];
    color2 = swapColors[color][1];
    color3 = swapColors[color][2];
    color4 = swapColors[color][3];


    // String index
    x = 0;
    // Store the caracter.
    character = text[x];

    while (character != '\0') {

        // If a spece or an unsupported char, the left an space.
        if (character != 32) { //32 = SPACE


            // EXCEPTIONS
            switch(character){
                // .
                case 46:
                    character=60;
                    break;

                // ,
                case 44:
                    character=61;
                    break;

                // " --> este caracter no se usa.
                case 34:
                    character=62;
                    break;

                // !
                case 33:
                    character=64;
                    break;

                // (
                case 40:
                    character=91;
                    break;

                // )
                case 41:
                    character=93;
                    break;

                // /
                case 47:
                    character=92;
                    break;
                // &
                case 38:
                    character=94;
                    break;
                // '
                case 39:
                    character=96;
                    break;
            }

            // I apply the offset.
            character-=48;

            // Create a copy of the char in a specific color:
            for(i=0;i<18;i++){
                pChar = font_chars + (character*18) + i;

                if(size==1) {
                    // NORMAL CHAR
                    if(*pChar == 0x55) colorchar[i]=color1;
                    else if(*pChar == 0xea) colorchar[i]=color2;
                    else if(*pChar == 0xd5) colorchar[i]=color3;
                    else if(*pChar == 0xff) colorchar[i]=color4;
                    else colorchar[i]=*pChar;

                } else {
                    // DOUBLE HEIGHT CHAR

                    pos=(i*size)-(i%FONT_W);

                    if(*pChar == 0x55) {
                        colorchar[pos]=color1;
                        colorchar[pos+FONT_W]=color1;
                    }
                    else if(*pChar == 0xea) {
                        colorchar[pos]=color2;
                        colorchar[pos+FONT_W]=color2;
                    }
                    else if(*pChar == 0xd5) {
                        colorchar[pos]=color3;
                        colorchar[pos+FONT_W]=color3;
                    }
                    else if(*pChar == 0xff) {
                        colorchar[pos]=color4;
                        colorchar[pos+FONT_W]=color4;

                    }
                    else {
                        colorchar[pos]=*pChar;
                        colorchar[pos+FONT_W]=*pChar;

                    }

                }

            }

            
            pvideo = cpct_getScreenPtr(CPCT_VMEM_START, xPos, yPos);
            if(transparent) cpct_drawSpriteMaskedAlignedTable(colorchar, pvideo, FONT_W, FONT_H*size, transparentMaskTable);
            else cpct_drawSprite (colorchar, pvideo, FONT_W, FONT_H*size);

            // Correction to support narrow wide, like "i", "l" and "'".
            if(character == 48 || character == 60 || character == 57) xPos--;
            
        }
        character = text[++x];
        xPos+=FONT_W;
    }
}


////////////////////////////////////////////////////////////////////////////////
//
// Draw a text centered on screen according to xMargin, 
// with a pause after draw every letter.
//

void DrawTextWithCadence(const u8 text[], u8 xMargin, u8 originY, u8 colorTxt) {

    u8 i=0, offsetX=0, offsetY=LINEHEIGHT, character[]="_";

    while (character[0] != '\0') {
        // Store the caracter.
        character[0] = text[i];

        if(offsetX==0 && character[0] == ' ') offsetX-=2;

        //cpct_waitVSYNC ();
        wait_n_VSYNCs(2);

        drawText(character, offsetX+xMargin, offsetY+originY, colorTxt, 1, 0);
        i++;

        // Correction to support narrow wide chars, like "i", "l" and "'".
        if(character[0] == 'i' || character[0] == 'l' || character[0] == '\'') --offsetX;

        offsetX+=2;
        if(offsetX>80-xMargin) {
            offsetX=0;
            offsetY+=LINEHEIGHT;
        }
    }
}

////////////////////////////////////////////////////////////////////////////////
//
// trace function for print numbers with debug purpouses...
//

/*
void traceNumber(u16 num, u8 x, u8 y){

    char text[6];  // 5 are the maximus digits suported by uitoa().

    uitoa(num, text);
    drawText(text, x, y, COLORTXT_WHITE, NORMALHEIGHT, OPAQUE);

}*/