//-----------------------------LICENSE NOTICE------------------------------------
//   ____        _           _       _____      _                                
//  |  _ \      | |         ( )     |  __ \    | |                               
//  | |_) | __ _| |__   __ _|/ ___  | |__) |_ _| | __ _  ___ ___                 
//  |  _ < / _` | '_ \ / _` | / __| |  ___/ _` | |/ _` |/ __/ _ |                
//  | |_) | (_| | |_) | (_| | \__ \ | |  | (_| | | (_| | (_|  __/                
//  |____/ \__,_|_.__/ \__,_| |___/ |_|   \__,_|_|\__,_|\___\___|                
//
//  Code & Gfx Copyright (C) 2017 Rafael Castillo (@Azicuetano)
//  Music & Fx Copyright (C) 2017 John McKlain (@elmcklain)
//  This file is part of BABA'S PALACE: an Amstrad CPC Game made with CPCTelera
//
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
//------------------------------------------------------------------------------

#include "text.h"
#include "../sprites/char.h"



 //Font Sprite Size
#define FONT_W 2
#define FONT_H 9


// COMPLETE CHARSET
// u8* const myFont[65] = {
//         font_char_00, font_char_01, font_char_02, font_char_03, font_char_04, font_char_05, font_char_06, font_char_07, font_char_08, font_char_09, font_char_10, font_char_11, font_char_12, font_char_13, font_char_14,
//         font_char_15, font_char_16, font_char_17, font_char_18, font_char_19, font_char_20, font_char_21, font_char_22, font_char_23, font_char_24, font_char_25, font_char_26, font_char_27, font_char_28, font_char_29,
//         font_char_30, font_char_31, font_char_32, font_char_33, font_char_34, font_char_35, font_char_36, font_char_37, font_char_38, font_char_39, font_char_40, font_char_41, font_char_42, font_char_43, font_char_44,
//         font_char_45, font_char_46, font_char_47, font_char_48, font_char_49, font_char_50, font_char_51, font_char_52, font_char_53, font_char_54, font_char_55, font_char_56, font_char_57, font_char_58, font_char_59,
//         font_char_60, font_char_61, font_char_62, font_char_63, font_char_64
// };

// ADAPTED CHARSET 
// 1. Strongly optimized in size.
// 2. Excluding chars not used in texts.
// 3. With a trick in the first date to print '.' when found ':' and do not perform a pause.
u8* const myFont[64] = {
        font_char_02,   0x4000,         font_char_02,   font_char_03,   0x4000,         font_char_05,   font_char_06,   font_char_07,   font_char_08,   0x4000,         font_char_10,   0x4000,         0x4000,         0x4000,         font_char_14,
        font_char_15,   0x4000,         font_char_17,   font_char_18,   0x4000,         0x4000,         0x4000,         font_char_22,   0x4000,         0x4000,         font_char_25,   0x4000,         0x4000,         0x4000,         font_char_29,
        0x4000,         font_char_31,	0x4000,         0x4000,         0x4000,         0x4000,         0x4000,         font_char_37,	font_char_38,	font_char_39,	font_char_40,	font_char_41,	font_char_42,	font_char_43,	font_char_44,
        font_char_45,	font_char_46,	font_char_47,	0x4000,         font_char_49,	font_char_50,	font_char_51,	font_char_52,	font_char_53,	font_char_54,	0x4000,         font_char_56,	font_char_57,	font_char_58,	font_char_59,
        font_char_60,	font_char_61,	font_char_62,   font_char_63
};


u8	const swapColors[2][4] = {
        {0x55, 0xea, 0xd5, 0xff},   // Bright White
        {0x15, 0x6a, 0x95, 0x3f}    // Bright Yellow
};

//
// PRINT SPRITE TEXT
//
// I've created a special character set with numbres, lowcases, uppercases and some puntuation symbols.
// All my charset are like Amstrad char set, but with some exceptions.
// 1) My 1sr char (char[0]) is Amstrad char[48], so I've to work with an offset of 48.
// 2) My charset has 75 chars.
// 3) Almost all chars are 4px wide, except 3 especial narrow who are 2px wide: ' i l
// 4) Some chars has not coincidence with Amstrad index because they are out of my 75 chars range. 
// These exceptions are: . , " ! ( ) & / '
// 
// color: to print chars in a special predefined color.
// size: To print normal chars (1) or double height chars (2).


void drawText(const u8 text[], u8 xPos, u8 yPos, u8 color) {

    u8* pvideo;
    u8 character;
    u8 x = 0; // general purpose.
    u8 i = 0; // chage color bucle.
    u8 color1=0, color2=0, color3=0, color4=0; // Store pair of pixel for print chars in a predefined color.
    u8 colorchar[2 * 9 * 2]; // Store a copy of a char in a specific color. // need double memory for double heigth chars.

    color1 = swapColors[color][0];
    color2 = swapColors[color][1];
    color3 = swapColors[color][2];
    color4 = swapColors[color][3];


    // String index
    x = 0;
    // Store the caracter.
    character = text[x];

    while (character != '\0') {

        // If a space or an unsupported char, the left an space.
        if (character != 32) //32 = SPACE
        { 

            // EXCEPTIONS
            if (character==46) character=60;
            if (character==44) character=61;
            if (character==33) character=64;
            if (character==39) character=96;

            // I apply the offset.
            //character-=48;
            character-=58;

            // Create a copy of the char in a specific color:
            for(i=0;i<18;i++){

                if(myFont[character][i]==0x55) colorchar[i]=color1;
                else if(myFont[character][i]==0xea) colorchar[i]=color2;
                else if(myFont[character][i]==0xd5) colorchar[i]=color3;
                else if(myFont[character][i]==0xff) colorchar[i]=color4;
                else colorchar[i]=myFont[character][i];
            }
            
            pvideo = cpct_getScreenPtr(CPCT_VMEM_START, xPos, yPos);
            cpct_drawSprite (colorchar, pvideo, FONT_W, FONT_H);

            // Correction to support narrow wide, like "i", "l" and "'".
            if(character == 38 || character == 50 || character == 47) xPos--;
          
        }
        character = text[++x];
        xPos+=FONT_W;
    }
}